/** This module handles stored options, currently all except input maps.
*
* The purpose of having all options centrally controlled is to allow generic handling by the GUI
* and ease saving and loading of values.
*/
module mde.options;

import mde.exception;

import mde.mergetag.read;
import mde.mergetag.write;
import mde.mergetag.dataset;
import mde.mergetag.exception;

import tango.scrapple.text.convert.parseTo : parseTo;
import tango.scrapple.text.convert.parseFrom : parseFrom;

import tango.io.FilePath : FilePath;
import tango.util.log.Log : Log, Logger;

Options options;
static this() {
    // Since options is used in many places, creating a class of default values now makes its use safer.
    options = new Options();
}

// Later: add automatic saving/loading for variables, variable lists with GUI name and description.

/** Class instance stores all options.
*
* All options and handling should be non-static to allow possibility of profiles later.
*/
class Options : DataSection
{
    /* The options.
    *
    * For now, named variables. Later, lists (and named variables?).
    *
    * They can be read and set directly by other code (so not thread-safe for writing currently).
    */
    char[] greeting;    // just a testing message
    char[] L10n;        // locale, e.g. en-GB
    
    /* The code to load/save the values.
    *
    * Uses mergetag.
    */
    void addTag (char[] tp, ID id, char[] dt) {
        if (tp == "char[]") {
            if (id == cast(ID)"greeting") greeting = parseTo!(char[]) (dt);
            else if (id == cast(ID)"L10n") L10n = parseTo!(char[]) (dt);
        }
    }
    void writeAll (ItemDelg dlg) {
        dlg ("char[]", cast(ID)"greeting", parseFrom!(char[]) (greeting));
        dlg ("char[]", cast(ID)"L10n", parseFrom!(char[]) (L10n));
    }
    
    
    /* Load/save options from file.
    *
    * If the file doesn't exist, no reading is attempted (options are left at default values).
    */
    private static const fileName = "conf/options.mtt";
    private static const secName = cast(ID)"Default";
    static void load () {
        FilePath filePath = new FilePath (fileName);
        // If it's not a file or an empty file stop. It should still be created on exit (assuming
        // it's not a folder).
        if (!filePath.exists || filePath.fileSize == 0u) return;
        
        Reader reader;
        try {
            reader = new Reader(filePath);
            reader.dataSecCreator = delegate DataSection(ID) {
                return new Options;
            };
            reader.read;
        } catch (MTException e) {
            logger.error ("Mergetag exception occurred:");
            logger.error (e.msg);
            throw new optionsLoadException ("Loading aborted: mergetag exception");
        }
        
        DataSection* secP = secName in reader.dataset.sec;
        options = cast(Options) *secP;
        if (options is null) {
            throw new optionsLoadException ("Loading failed: expected section not found");
        }
        // else loading was succesful.
    }
    static void save () {
        DataSet ds = new DataSet();
        ds.sec[secName] = options;
        
        IWriter writer;
        try {
            writer = makeWriter (fileName, ds);
            writer.write();
        } catch (MTException e) {
            logger.error ("Mergetag exception occurred; saving aborted:");
            logger.error (e.msg);
            //FIXME: currently nothing done besides logging the error
        }
    }
    
    private static Logger logger;
    static this() {
        logger = Log.getLogger ("mde.options");
    }
}
